/*
 * BasicDelegate.hh
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file BasicDelegate.hh
 *
 * Header file for a delegate that makes basic reports to a
 * Broker::RealTimeAdvocate.
 */

#ifndef _basic_delegate_hh
#define _basic_delegate_hh

#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <rk.h>
#include <rk_stub.h>

#include <time_util.h>

#include <BrokerC.h>

#include "FakeDelegate.hh"

/**
 * A simple delegate that computes the number of deadlines missed and passes
 * the value to Broker::RealTimeAdvocate::ReportCPU.
 */
class BasicDelegate : public FakeDelegate
{
    
public:

    /**
     * Construct a BasicDelegate with the given values.  This includes:
     *
     * @li Creating the process with rk_stub_mk_pid.
     * @li Initializing the bd_TaskDescription and bd_CPUSchedule parameter
     *     lists.
     * @li and initializing internal state.
     *
     * @param name The name of the process.  This will be passed on to
     *             rk_stub_mk_pid.
     * @param period The period of the process in microseconds.
     * @param deadline The deadline of the process in microseconds and relative
     *                 to the start of the period.
     */
    BasicDelegate(const char *name,
		  unsigned long period,
		  unsigned long deadline);

    /**
     * Deconstruct a BasicDelegate.
     */
    virtual ~BasicDelegate();

    /**
     * Premethod callback that just records the current time.
     *
     * @return RKSP_OK if the process is on schedule or RKSP_DROP if it is
     *         running late and needs to drop data in order to catch up.
     */
    virtual rk_stub_precall_retval_t precall(void);

    /**
     * Postmethod callback that computes how many deadlines have been missed
     * and reports the results to bd_Advocate.
     */
    virtual void postcall(void);

    /**
     * TaskParameters constructed by the object that can be passed on to
     * Broker::TaskFactory::CreateTask.
     */
    Broker::TaskParameters bd_TaskDescription;

    /**
     * ScheduleParameters constructed by the object that can be passed on to
     * Broker::Task::BeginScheduling.
     */
    Broker::ScheduleParameters bd_CPUSchedule;

    /**
     * The process' advocate, this must be set by the code using the object.
     * The user must also call Broker::Task::BeginCPUScheduling and
     * Broker::Task::EndCPUScheduling.
     */
    Broker::RealTimeTask_var bd_Advocate;

protected:

    /**
     * The period of the process in microseconds.
     */
    unsigned long bd_Period;
    
    /**
     * The deadline of the process in microseconds and relative to the start of
     * the period.
     */
    unsigned long bd_Deadline;

    /**
     * The number of data segments to drop in order to get back on schedule.
     * This will be set by postcall() when it detects how many deadlines have
     * been missed.
     */
    int bd_AdviseDrop;

    /**
     * The simulated process ID.
     */
    pid_t bd_PID;

    /**
     * The "method" starting time as recorded by precall().
     */
    struct timespec bd_Start;

    /**
     * The resource usage at the "method" starting time.
     */
    struct rusage bd_RUStart;
    
};

#endif
