#! /bin/awk

#
# rktimes_sum.awk
#
# Copyright (c) 2003 The University of Utah and the Flux Group.
# All rights reserved.
#
# This file is licensed under the terms of the GNU Public License.  
# See the file "license.terms" for restrictions on redistribution 
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

#
# rktimes_offset.awk - Offset the X values in one data file based on another.
#
# Usage: awk -f rktimes_offset.awk <rktimes-child> <file-to-align>
#
# The rktimes_offset.awk script is used to align the data points in one file
# with those in another.  Alignment might be needed if one file was recorded at
# a later time than the other.  The script works by computing the difference
# between the start time for each file and adjusting the X axis for the second
# file accordingly.  The result can then be displayed simultaneously with the
# first file without any alignment problems.
#
# Example:
#
#   To offset the foo.out file based on the rktimes-1234.out file.
#
#     [rktimes-1234.out]
#     # start: 20
#     0 1.0
#     1 1.5
#     ...
#
#     [foo.out]
#     # start: 22
#     0 2.0
#     1 3.0
#     ...
#
#     $ awk -f rktimes_offset.awk rktimes-1234.out foo.out
#     # start: 20
#     2 2.0
#     3 3.0
#     ...
#
# See Also:
#
#   rktimes(1), rktimes_sum.awk
#

BEGIN {
    if( ARGC != 3 )
    {
	printf("Usage: awk -f rktimes_offset.awk <rktimes-child> <file-to-align>\n") > "/dev/stderr";
	exit 1;
    }
    sourcing = 1;
    printing = 0;
}

# Match the start time comment.
/^\# start:/ {
    if( sourcing )
    {
	start_time = $3;
	ARGIND += 1;
	sourcing = 0;
    }
    else
    {
	diff_time = $3 - start_time;
	printing = 1;
	printf("# start: %f\n", start_time);
    }
}

# Match a data point and offset the X value.
/[0-9\.\-]+ [0-9\.\-]+/ {
    if( printing )
    {
	printf("%f %f\n", $1 + diff_time, $2);
    }
}
