/*
 * LoggingAdvocate.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file LoggingAdvocate.cc
 *
 * Implementation of the LoggingAdvocate class.
 */

#include "config.h"

#include <string.h>
#include <sys/time.h>

#include <iostream>
#include <fstream>
#include <iomanip>

#include "LoggingAdvocate.hh"

LoggingAdvocate::LoggingAdvocate(void)
{
    gettimeofday(&this->eta_StartTime, NULL);
    this->eta_LastCPULog = this->eta_StartTime;
    this->eta_CPULog = NULL;
    this->eta_LastAdviceLog = this->eta_StartTime;
    this->eta_AdviceLog = NULL;
    this->eta_CompleteLog = NULL;
}

LoggingAdvocate::~LoggingAdvocate()
{
    if( this->eta_CPULog != NULL )
	delete this->eta_CPULog;
    if( this->eta_AdviceLog != NULL )
	delete this->eta_AdviceLog;
    if( this->eta_CompleteLog != NULL )
	delete this->eta_CompleteLog;
}

void LoggingAdvocate::SetDelegateAttribute(const char *name,
					   const CORBA::Any &value)
    throw (CORBA::SystemException)
{
    if( (strcasecmp(name, "cpu_log") == 0) ||
	(strcasecmp(name, "cpu-log") == 0) ||
	(strcasecmp(name, "cpu log") == 0) )
    {
	const char *cpu_log;
	
	if( this->eta_CPULog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("cpu log");
	}
	else if( value >>= cpu_log )
	{
	    this->eta_CPULog = new ofstream(cpu_log);
	    (*this->eta_CPULog) << setfill('0');
	    (*this->eta_CPULog)
		<< "# start: "
		<< (long)this->eta_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->eta_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "advice_log") == 0) ||
	     (strcasecmp(name, "advice-log") == 0) ||
	     (strcasecmp(name, "advice log") == 0) )
    {
	const char *advice_log;
	
	if( this->eta_AdviceLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("advice log");
	}
	else if( value >>= advice_log )
	{
	    this->eta_AdviceLog = new ofstream(advice_log);
	    (*this->eta_AdviceLog) << setfill('0');
	    (*this->eta_AdviceLog)
		<< "# start: "
		<< (long)this->eta_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->eta_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else if( (strcasecmp(name, "complete_log") == 0) ||
	     (strcasecmp(name, "complete-log") == 0) ||
	     (strcasecmp(name, "complete log") == 0) )
    {
	const char *complete_log;
	
	if( this->eta_CompleteLog != NULL )
	{
	    throw Broker::DuplicateTaskParameter("complete log");
	}
	else if( value >>= complete_log )
	{
	    this->eta_CompleteLog = new ofstream(complete_log);
	    (*this->eta_CompleteLog) << setfill('0');
	    (*this->eta_CompleteLog)
		<< "# start: "
		<< (long)this->eta_StartTime.tv_sec
		<< "."
		<< setw(6) << (long)this->eta_StartTime.tv_usec
		<< endl;
	}
	else
	{
	    throw CORBA::BAD_PARAM();
	}
    }
    else
    {
	this->RealTimeTaskDelegateImpl::SetDelegateAttribute(name, value);
    }
}

void LoggingAdvocate::ReportCPU(Broker::RealTimeTask_ptr rtt,
				CORBA::ULong status,
				CORBA::ULong advice)
    throw (CORBA::SystemException)
{
    if( CORBA::is_nil(this->dm_RemoteObject) )
    {
	throw CORBA::BAD_INV_ORDER();
    }
    
    if( this->eta_CompleteLog != NULL )
    {
	struct timeval tv, diff;
	
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->eta_StartTime, &diff);
	(*this->eta_CompleteLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " -0.01"
	    << endl;
    }
    if( this->eta_CPULog != NULL )
    {
	struct timeval tv, st, diff;
	
	timersub(&this->eta_LastCPULog, &this->eta_StartTime, &diff);
	st.tv_sec = status / 1000000;
	st.tv_usec = status % 1000000;
	(*this->eta_CPULog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->eta_StartTime, &diff);
	(*this->eta_CPULog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
	this->eta_LastCPULog = tv;
    }
    if( this->eta_AdviceLog != NULL )
    {
	struct timeval tv, st, diff;

	st.tv_sec = advice / 1000000;
	st.tv_usec = advice % 1000000;
#if 0
	timersub(&this->eta_LastAdviceLog, &this->eta_StartTime, &diff);
	(*this->eta_AdviceLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
#endif
	gettimeofday(&tv, NULL);
	timersub(&tv, &this->eta_StartTime, &diff);
	(*this->eta_AdviceLog)
	    << (long)diff.tv_sec
	    << "."
	    << setw(6) << (long)diff.tv_usec
	    << " "
	    << (long)st.tv_sec
	    << "."
	    << setw(6) << (long)st.tv_usec
	    << endl;
	this->eta_LastAdviceLog = tv;
    }
    this->RealTimeTaskDelegateImpl::ReportCPU(rtt, status, advice);
}
