/*
 * TaskFactory.cc
 *
 * Copyright (c) 2003 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

/**
 * @file TaskFactory.cc
 *
 * The shared library interface for libtask_factories.
 */

#include "config.h"

#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

#include <iostream>

#include <assert_pp.h>
#include <HeyParser.hh>
#include <factory_library.h>

#if defined(HAVE_QUO)
#include "QuoAdvocate.hh"
#endif

#include "LoggingAdvocate.hh"
#include "GlacialTaskAdvocate.hh"
#include "MaxDecayTaskAdvocate.hh"

using namespace std;

/**
 * Handle 'hey's directed at the library.
 *
 * @param orb Pointer to the ORB.
 * @param hp The HeyParser that describes the action to take.
 * @param out Standard out.
 * @param err Standard error.
 * @return Zero on success, an errno value otherwise.
 */
static int tfHey(CORBA::ORB_ptr orb, HeyParser &hp, ostream &out, ostream &err)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("logging-advocate",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"Create a LoggingAdvocate object.\n"),
	HeyPropertyInfo("max-decay-task-advocate",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"Create a MaxDecayTaskAdvocate object.\n"),
	HeyPropertyInfo("glacial-task-advocate",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"Create a GlacialTaskAdvocate object.\n"),
#if defined(HAVE_QUO)
	HeyPropertyInfo("quo-advocate",
			(1L << HeyParser::CREATE_PROPERTY),
			"",
			"Create a QuoAdvocate object.\n"),
#endif
	HeyPropertyInfo::HPI_NULL
    };

    const char *prop_name, *prop_value;
    int retval = EXIT_FAILURE;

    require(!CORBA::is_nil(orb));

    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	}
	catch(const HeyParserException &e)
	{
	    out << "logging-advocate" << endl;
	    out << "max-decay-task-advocate" << endl;
	    out << "glacial-task-advocate" << endl;
#if defined(HAVE_QUO)
	    out << "quo-advocate" << endl;
#endif
	}
	break;
    case HeyParser::CREATE_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( strcasecmp(prop_name, "logging-advocate") == 0 )
	{
	    LoggingAdvocate *la = new LoggingAdvocate();
	    BrokerDelegates::RealTimeTaskDelegate_var rttd;
	    
	    rttd = la->_this();
	    out << orb->object_to_string(rttd._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
	else if( strcasecmp(prop_name, "max-decay-task-advocate") == 0 )
	{
	    MaxDecayTaskAdvocate *mdta = new MaxDecayTaskAdvocate();
	    BrokerDelegates::RealTimeTaskDelegate_var rttd;
	    
	    rttd = mdta->_this();
	    out << orb->object_to_string(rttd._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
	else if( strcasecmp(prop_name, "glacial-task-advocate") == 0 )
	{
	    GlacialTaskAdvocate *gta = new GlacialTaskAdvocate();
	    BrokerDelegates::RealTimeTaskDelegate_var rttd;
	    
	    rttd = gta->_this();
	    out << orb->object_to_string(rttd._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
#if defined(HAVE_QUO)
	else if( strcasecmp(prop_name, "quo-advocate") == 0 )
	{
	    QuoAdvocate *qa = new QuoAdvocate();
	    BrokerDelegates::RealTimeTaskDelegate_var rttd;
	    
	    rttd = qa->_this();
	    out << orb->object_to_string(rttd._retn()) << endl;
	    retval = EXIT_SUCCESS;
	}
#endif
	else
	{
	    err << "Unknown property: " << prop_name << endl;
	}
	break;
    case HeyParser::GET_SUITES:
	out << suites;
	retval = EXIT_SUCCESS;
	break;
    }
    return( retval );
}

#if defined(HAVE_QUO)
/**
 * Handle 'hey's directed at an IMP:QuoAdvocate:1.0 object.
 *
 * @param orb Pointer to the ORB.
 * @param hp The HeyParser that describes the action to take.
 * @return Zero on success, an errno value otherwise.
 */
static int tfQuoAdvocateHey(CORBA::ORB_ptr orb, HeyParser &hp)
{
    static HeyPropertyInfo suites[] = {
	HeyPropertyInfo("status-sc",
			(1L << HeyParser::GET_PROPERTY) |
			(1L << HeyParser::SET_PROPERTY),
			"",
			"Get or set the QuO system condition that records the status given to ReportCPU.\n"),
	HeyPropertyInfo("advice-sc",
			(1L << HeyParser::GET_PROPERTY) |
			(1L << HeyParser::SET_PROPERTY),
			"",
			"Get or set the QuO system condition that records the advice given to ReportCPU.\n"),
	HeyPropertyInfo::HPI_NULL
    };
    
    BrokerDelegates::Delegate_var delegate;
    const char *prop_name, *prop_value;
    int retval = EXIT_FAILURE;
    CORBA::Object_var obj;
    
    obj = orb->string_to_object(hp.who());
    delegate = BrokerDelegates::Delegate::_narrow(obj.in());
    if( CORBA::is_nil(delegate.in()) )
    {
	throw CORBA::BAD_PARAM();
    }
    
    switch( hp.what() )
    {
    case HeyParser::LIST_PROPERTIES:
	try
	{
	    hp.popProperty(prop_name, prop_value);
	}
	catch(const HeyParserException &e)
	{
	    cout << "status-sc" << endl;
	    cout << "advice-sc" << endl;
	    retval = EXIT_SUCCESS;
	}
	break;
    case HeyParser::GET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( (strcasecmp(prop_name, "status-sc") == 0) ||
	    (strcasecmp(prop_name, "advice-sc") == 0) )
	{
	    quo::ValueSC_ptr scp;
	    CORBA::Any_ptr any;

	    any = delegate->GetDelegateAttribute(prop_name);
	    if( (*any) >>= scp )
	    {
		cout << orb->object_to_string(scp) << endl;
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
		cerr << "Unknown error..." << endl;
	    }
	}
	else
	{
	    cerr << "Unknown property: " << prop_name;
	}
	break;
    case HeyParser::SET_PROPERTY:
	hp.popProperty(prop_name, prop_value);
	if( (strcasecmp(prop_name, "status-sc") == 0) ||
	    (strcasecmp(prop_name, "advice-sc") == 0) )
	{
	    quo::ValueSC_var sc;
	    CORBA::Any any;

	    obj = orb->string_to_object(hp.to());
	    sc = quo::ValueSC::_narrow(obj.in());
	    if( CORBA::is_nil(sc.in()) )
	    {
		cerr << "Error: IOR is not a system condition" << endl;
	    }
	    else
	    {
		any <<= sc;
		delegate->SetDelegateAttribute(prop_name, any);
		cout << "ok" << endl;
		retval = EXIT_SUCCESS;
	    }
	}
	else
	{
	    cerr << "Unknown property: " << prop_name;
	}
	break;
    case HeyParser::GET_SUITES:
	cout << suites;
	retval = EXIT_SUCCESS;
	break;
    default:
	cerr << "Unknown verb..." << endl;
	break;
    }
    return( retval );
}
#endif

int FACTORY_METHOD_SYMBOL(factory_library_op_t op, int tag, va_list args)
{
    int retval = EXIT_FAILURE;

    struct {
	const char *hey_server_hint;
	HeyParser *hey_parser;
	CORBA::ORB_ptr orb;
	ostream *out;
	ostream *err;
    } attr = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    };
    
    while( tag != FMA_TAG_DONE )
    {
	switch( tag )
	{
	case FMA_ORB:
	    attr.orb = va_arg(args, CORBA::ORB_ptr);
	    break;
	case FMA_HeyParser:
	    attr.hey_parser = va_arg(args, HeyParser *);
	    break;
	case FMA_HeyServerHint:
	    attr.hey_server_hint = va_arg(args, const char *);
	    break;
	case FMA_StandardOut:
	    attr.out = va_arg(args, ostream *);
	    break;
	case FMA_StandardError:
	    attr.err = va_arg(args, ostream *);
	    break;
	default:
	    break;
	}
	tag = va_arg(args, int);
    }

    switch( op )
    {
    case FLO_QUERY:
#if defined(HAVE_QUO)
	if( attr.hey_server_hint != NULL )
	{
	    if( strcmp(attr.hey_server_hint, "IMP:QuoAdvocate:1.0") == 0 )
	    {
		retval = EXIT_SUCCESS;
	    }
	    else
	    {
	    }
	}
	else
	{
	    cout << "\tIMP:QuoAdvocate:1.0" << endl;
	}
#endif
	break;
    case FLO_HEY:
	if( attr.hey_server_hint != NULL )
	{
#if defined(HAVE_QUO)
	    try
	    {
		if( strcmp(attr.hey_server_hint,
			   "IMP:QuoAdvocate:1.0") == 0 )
		{
		    retval = tfQuoAdvocateHey(attr.orb, *attr.hey_parser);
		}
		else
		{
		}
	    }
	    catch(const HeyParserException &he)
	    {
		cerr << "Parse error: " << he << endl;
	    }
	    catch(const CORBA::SystemException &e)
	    {
		cerr << "Caught Exception: " << e << endl;
	    }
	    catch(...)
	    {
		cerr << "Caught an unhandled exception" << endl;
	    }
#endif
	}
	else
	{
	    retval = tfHey(attr.orb, *attr.hey_parser, *attr.out, *attr.err);
	}
	break;
    default:
	break;
    }
    
    return( retval );
}
